@@ -10,6 +10,7 @@
 #include "xs_time.h"
 #include "xs_set.h"
 #include "xs_match.h"
+#include "xs_unicode.h"
 
 #include "snac.h"
 
@@ -178,6 +179,11 @@ const char *get_atto(const xs_dict *msg)
             }
         }
     }
+    else
+    if (xs_type(actor) == XSTYPE_DICT) {
+        /* bandwagon.fm returns this */
+        actor = xs_dict_get(actor, "id");
+    }
 
     return actor;
 }
@@ -258,6 +264,10 @@ xs_list *get_attachments(const xs_dict *msg)
             d = xs_dict_append(d, "href", href);
             d = xs_dict_append(d, "name", name);
 
+            const xs_dict *icon = xs_dict_get(v, "icon");
+            if (xs_type(icon) == XSTYPE_DICT)
+                d = xs_dict_append(d, "icon", icon);
+
             l = xs_list_append(l, d);
         }
     }
@@ -697,6 +707,32 @@ int is_msg_for_me(snac *snac, const xs_dict *c_msg)
         }
     }
 
+    /* does this message contain a tag we are following? */
+    const xs_list *fw_tags = xs_dict_get(snac->config, "followed_hashtags");
+    if (pub_msg && xs_type(fw_tags) == XSTYPE_LIST) {
+        const xs_list *tags_in_msg = xs_dict_get(msg, "tag");
+        if (xs_type(tags_in_msg) == XSTYPE_LIST) {
+            const xs_dict *te;
+
+            /* iterate the tags in the message */
+            xs_list_foreach(tags_in_msg, te) {
+                if (xs_type(te) == XSTYPE_DICT) {
+                    const char *type = xs_dict_get(te, "type");
+                    const char *name = xs_dict_get(te, "name");
+
+                    if (xs_type(type) == XSTYPE_STRING && xs_type(name) == XSTYPE_STRING) {
+                        if (strcmp(type, "Hashtag") == 0) {
+                            xs *lc_name = xs_utf8_to_lower(name);
+
+                            if (xs_list_in(fw_tags, lc_name) != -1)
+                                return 7;
+                        }
+                    }
+                }
+            }
+        }
+    }
+
     return 0;
 }
 
@@ -1038,15 +1074,14 @@ xs_dict *msg_base(snac *snac, const char *type, const char *id,
 }
 
 
-xs_dict *msg_collection(snac *snac, const char *id)
+xs_dict *msg_collection(snac *snac, const char *id, int items)
 /* creates an empty OrderedCollection message */
 {
     xs_dict *msg = msg_base(snac, "OrderedCollection", id, NULL, NULL, NULL);
-    xs *ol = xs_list_new();
+    xs *n = xs_number_new(items);
 
     msg = xs_dict_append(msg, "attributedTo", snac->actor);
-    msg = xs_dict_append(msg, "orderedItems", ol);
-    msg = xs_dict_append(msg, "totalItems",   xs_stock(0));
+    msg = xs_dict_append(msg, "totalItems",   n);
 
     return msg;
 }
@@ -1387,8 +1422,10 @@ xs_dict *msg_follow(snac *snac, const char *q)
 
 
 xs_dict *msg_note(snac *snac, const xs_str *content, const xs_val *rcpts,
-                  const xs_str *in_reply_to, const xs_list *attach, int priv)
+                  const xs_str *in_reply_to, const xs_list *attach,
+                  int scope, const char *lang_str)
 /* creates a 'Note' message */
+/* scope: 0, public; 1, private (mentioned only); 2, "quiet public"; 3, followers only */
 {
     xs *ntid = tid(0);
     xs *id   = xs_fmt("%s/p/%s", snac->actor, ntid);
@@ -1404,6 +1441,9 @@ xs_dict *msg_note(snac *snac, const xs_str *content, const xs_val *rcpts,
     xs_list *p;
     const xs_val *v;
 
+    /* FIXME: implement scopes 2 and 3 */
+    int priv = scope == 1;
+
     if (rcpts == NULL)
         to = xs_list_new();
     else {
@@ -1477,20 +1517,31 @@ xs_dict *msg_note(snac *snac, const xs_str *content, const xs_val *rcpts,
 
     /* create the attachment list, if there are any */
     if (!xs_is_null(attach)) {
-        int c = 0;
-        while (xs_list_next(attach, &v, &c)) {
-            xs *d            = xs_dict_new();
+        xs_list_foreach(attach, v) {
             const char *url  = xs_list_get(v, 0);
             const char *alt  = xs_list_get(v, 1);
             const char *mime = xs_mime_by_ext(url);
+            int add = 1;
 
-            d = xs_dict_append(d, "mediaType", mime);
-            d = xs_dict_append(d, "url",       url);
-            d = xs_dict_append(d, "name",      alt);
-            d = xs_dict_append(d, "type",
-                xs_startswith(mime, "image/") ? "Image" : "Document");
+            /* check if it's already here */
+            const xs_dict *ad;
+            xs_list_foreach(atls, ad) {
+                if (strcmp(xs_dict_get_def(ad, "url", ""), url) == 0) {
+                    add = 0;
+                    break;
+                }
+            }
 
-            atls = xs_list_append(atls, d);
+            if (add) {
+                xs *d = xs_dict_new();
+                d = xs_dict_append(d, "mediaType", mime);
+                d = xs_dict_append(d, "url",       url);
+                d = xs_dict_append(d, "name",      alt);
+                d = xs_dict_append(d, "type",
+                    xs_startswith(mime, "image/") ? "Image" : "Document");
+
+                atls = xs_list_append(atls, d);
+            }
         }
     }
 
@@ -1538,6 +1589,20 @@ xs_dict *msg_note(snac *snac, const xs_str *content, const xs_val *rcpts,
     if (xs_list_len(atls))
         msg = xs_dict_append(msg, "attachment", atls);
 
+    /* set language content map */
+    if (xs_type(lang_str) == XSTYPE_STRING) {
+        /* split at the first _ */
+        xs *l0 = xs_split(lang_str, "_");
+        const char *lang = xs_list_get(l0, 0);
+
+        if (xs_type(lang) == XSTYPE_STRING && strlen(lang) == 2) {
+            /* a valid ISO language id */
+            xs *cmap = xs_dict_new();
+            cmap = xs_dict_set(cmap, lang, xs_dict_get(msg, "content"));
+            msg = xs_dict_set(msg, "contentMap", cmap);
+        }
+    }
+
     return msg;
 }
 
@@ -1579,7 +1644,7 @@ xs_dict *msg_question(snac *user, const char *content, xs_list *attach,
                       const xs_list *opts, int multiple, int end_secs)
 /* creates a Question message */
 {
-    xs_dict *msg = msg_note(user, content, NULL, NULL, attach, 0);
+    xs_dict *msg = msg_note(user, content, NULL, NULL, attach, 0, NULL);
     int max      = 8;
     xs_set seen;
 
@@ -2372,22 +2437,28 @@ void process_user_queue_item(snac *snac, xs_dict *q_item)
         xs *rcpts = recipient_list(snac, msg, 1);
         xs_set inboxes;
         const xs_str *actor;
-        int c;
 
         xs_set_init(&inboxes);
 
-        /* iterate the recipients */
-        c = 0;
-        while (xs_list_next(rcpts, &actor, &c)) {
-            xs *inbox = get_actor_inbox(actor, 1);
+        /* add this shared inbox first */
+        xs *this_shared_inbox = xs_fmt("%s/shared-inbox", srv_baseurl);
+        xs_set_add(&inboxes, this_shared_inbox);
+        enqueue_output(snac, msg, this_shared_inbox, 0, 0);
 
-            if (inbox != NULL) {
-                /* add to the set and, if it's not there, send message */
-                if (xs_set_add(&inboxes, inbox) == 1)
-                    enqueue_output(snac, msg, inbox, 0, 0);
+        /* iterate the recipients */
+        xs_list_foreach(rcpts, actor) {
+            /* local users were served by this_shared_inbox */
+            if (!xs_startswith(actor, srv_baseurl)) {
+                xs *inbox = get_actor_inbox(actor, 1);
+
+                if (inbox != NULL) {
+                    /* add to the set and, if it's not there, send message */
+                    if (xs_set_add(&inboxes, inbox) == 1)
+                        enqueue_output(snac, msg, inbox, 0, 0);
+                }
+                else
+                    snac_log(snac, xs_fmt("cannot find inbox for %s", actor));
             }
-            else
-                snac_log(snac, xs_fmt("cannot find inbox for %s", actor));
         }
 
         /* if it's a public note or question, send to the collected inboxes */
@@ -2396,8 +2467,7 @@ void process_user_queue_item(snac *snac, xs_dict *q_item)
                 xs *shibx = inbox_list();
                 const xs_str *inbox;
 
-                c = 0;
-                while (xs_list_next(shibx, &inbox, &c)) {
+                xs_list_foreach(shibx, inbox) {
                     if (xs_set_add(&inboxes, inbox) == 1)
                         enqueue_output(snac, msg, inbox, 0, 0);
                 }
@@ -2837,6 +2907,8 @@ int activitypub_get_handler(const xs_dict *req, const char *q_path,
 
     *ctype  = "application/activity+json";
 
+    int show_contact_metrics = xs_is_true(xs_dict_get(snac.config, "show_contact_metrics"));
+
     if (p_path == NULL) {
         /* if there was no component after the user, it's an actor request */
         msg = msg_actor(&snac);
@@ -2850,14 +2922,13 @@ int activitypub_get_handler(const xs_dict *req, const char *q_path,
     if (strcmp(p_path, "outbox") == 0 || strcmp(p_path, "featured") == 0) {
         xs *id = xs_fmt("%s/%s", snac.actor, p_path);
         xs *list = xs_list_new();
-        msg = msg_collection(&snac, id);
         const char *v;
-        int tc = 0;
+        int cnt = xs_number_get(xs_dict_get_def(srv_config, "max_public_entries", "20"));
 
         /* get the public outbox or the pinned list */
-        xs *elems = *p_path == 'o' ? timeline_simple_list(&snac, "public", 0, 20) : pinned_list(&snac);
+        xs *elems = *p_path == 'o' ? timeline_simple_list(&snac, "public", 0, cnt) : pinned_list(&snac);
 
-        while (xs_list_next(elems, &v, &tc)) {
+        xs_list_foreach(elems, v) {
             xs *i = NULL;
 
             if (valid_status(object_get_by_md5(v, &i))) {
@@ -2872,14 +2943,32 @@ int activitypub_get_handler(const xs_dict *req, const char *q_path,
         }
 
         /* replace the 'orderedItems' with the latest posts */
-        xs *items = xs_number_new(xs_list_len(list));
+        msg = msg_collection(&snac, id, xs_list_len(list));
         msg = xs_dict_set(msg, "orderedItems", list);
-        msg = xs_dict_set(msg, "totalItems",   items);
     }
     else
-    if (strcmp(p_path, "followers") == 0 || strcmp(p_path, "following") == 0) {
+    if (strcmp(p_path, "followers") == 0) {
+        int total = 0;
+
+        if (show_contact_metrics) {
+            xs *l = follower_list(&snac);
+            total = xs_list_len(l);
+        }
+
         xs *id = xs_fmt("%s/%s", snac.actor, p_path);
-        msg = msg_collection(&snac, id);
+        msg = msg_collection(&snac, id, total);
+    }
+    else
+    if (strcmp(p_path, "following") == 0) {
+        int total = 0;
+
+        if (show_contact_metrics) {
+            xs *l = following_list(&snac);
+            total = xs_list_len(l);
+        }
+
+        xs *id = xs_fmt("%s/%s", snac.actor, p_path);
+        msg = msg_collection(&snac, id, total);
     }
     else
     if (xs_startswith(p_path, "p/")) {
