@@ -6,6 +6,7 @@
 #include "xs.h"
 #include "xs_hex.h"
 #include "xs_io.h"
+#include "xs_unicode_tbl.h"
 #include "xs_unicode.h"
 #include "xs_json.h"
 #include "xs_curl.h"
@@ -32,6 +33,7 @@
 xs_str *srv_basedir = NULL;
 xs_dict *srv_config = NULL;
 xs_str *srv_baseurl = NULL;
+xs_str *srv_proxy_token_seed = NULL;
 
 int dbglevel = 0;
 
