@@ -55,7 +55,7 @@ static const char *default_css =
     ".snac-author-tag { font-size: 80% }\n"
     ".snac-pubdate { color: #a0a0a0; font-size: 90% }\n"
     ".snac-top-controls { padding-bottom: 1.5em }\n"
-    ".snac-post { border-top: 1px solid #a0a0a0; }\n"
+    ".snac-post { border-top: 1px solid #a0a0a0; padding-top: 0.5em; padding-bottom: 0.5em; }\n"
     ".snac-children { padding-left: 1em; border-left: 1px solid #a0a0a0; }\n"
     ".snac-textarea { font-family: inherit; width: 100% }\n"
     ".snac-history { border: 1px solid #606060; border-radius: 3px; margin: 2.5em 0; padding: 0 2em }\n"
@@ -69,6 +69,9 @@ static const char *default_css =
     ".snac-btn-unlimit { float: right; margin-left: 0.5em }\n"
     ".snac-footer { margin-top: 2em; font-size: 75% }\n"
     ".snac-poll-result { margin-left: auto; margin-right: auto; }\n"
+    ".snac-list-of-lists { padding-left: 0; }\n"
+    ".snac-list-of-lists li { display: inline; border: 1px solid #a0a0a0; border-radius: 25px;\n"
+    "  margin-right: 0.5em; padding-left: 0.5em; padding-right: 0.5em; }\n"
     "@media (prefers-color-scheme: dark) { \n"
     "  body, input, textarea { background-color: #000; color: #fff; }\n"
     "  a { color: #7799dd }\n"
@@ -104,6 +107,22 @@ static const char *greeting_html =
     "<p>This site is powered by <abbr title=\"Social Networks Are Crap\">snac</abbr>.</p>\n"
     "</body></html>\n";
 
+
+int write_default_css(void)
+{
+    FILE *f;
+
+    xs *sfn = xs_fmt("%s/style.css", srv_basedir);
+    if ((f = fopen(sfn, "w")) == NULL)
+        return 1;
+
+    fwrite(default_css, strlen(default_css), 1, f);
+    fclose(f);
+
+    return 0;
+}
+
+
 int snac_init(const char *basedir)
 {
     FILE *f;
@@ -214,15 +233,11 @@ int snac_init(const char *basedir)
     fwrite(gh, strlen(gh), 1, f);
     fclose(f);
 
-    xs *sfn = xs_fmt("%s/style.css", srv_basedir);
-    if ((f = fopen(sfn, "w")) == NULL) {
-        printf("ERROR: cannot create '%s'\n", sfn);
+    if (write_default_css()) {
+        printf("ERROR: cannot create style.css\n");
         return 1;
     }
 
-    fwrite(default_css, strlen(default_css), 1, f);
-    fclose(f);
-
     xs *cfn = xs_fmt("%s/server.json", srv_basedir);
     if ((f = fopen(cfn, "w")) == NULL) {
         printf("ERROR: cannot create '%s'\n", cfn);
@@ -431,7 +446,8 @@ int deluser(snac *user)
 void verify_links(snac *user)
 /* verifies a user's links */
 {
-    const xs_dict *p = xs_dict_get(user->config, "metadata");
+    xs *metadata = NULL;
+    const xs_dict *md = xs_dict_get(user->config, "metadata");
     const char *k, *v;
     int changed = 0;
 
@@ -439,8 +455,30 @@ void verify_links(snac *user)
     headers = xs_dict_append(headers, "accept", "text/html");
     headers = xs_dict_append(headers, "user-agent", USER_AGENT " (link verify)");
 
+    if (xs_type(md) == XSTYPE_DICT)
+        metadata = xs_dup(md);
+    else
+    if (xs_type(md) == XSTYPE_STRING) {
+        /* convert to dict for easier iteration */
+        metadata = xs_dict_new();
+        xs *l = xs_split(md, "\n");
+        const char *ll;
+
+        xs_list_foreach(l, ll) {
+            xs *kv = xs_split_n(ll, "=", 1);
+            const char *k = xs_list_get(kv, 0);
+            const char *v = xs_list_get(kv, 1);
+
+            if (k && v) {
+                xs *kk = xs_strip_i(xs_dup(k));
+                xs *vv = xs_strip_i(xs_dup(v));
+                metadata = xs_dict_set(metadata, kk, vv);
+            }
+        }
+    }
+
     int c = 0;
-    while (p && xs_dict_next(p, &k, &v, &c)) {
+    while (metadata && xs_dict_next(metadata, &k, &v, &c)) {
         /* not an https link? skip */
         if (!xs_startswith(v, "https:/" "/"))
             continue;
@@ -550,3 +588,319 @@ void verify_links(snac *user)
             rename(bfn, fn);
     }
 }
+
+
+void export_csv(snac *user)
+/* exports user data to current directory in a way that pleases Mastodon */
+{
+    FILE *f;
+    xs *fn = NULL;
+
+    fn = xs_fmt("%s/export/bookmarks.csv", user->basedir);
+    if ((f = fopen(fn, "w")) != NULL) {
+        snac_log(user, xs_fmt("Creating %s...", fn));
+
+        xs *l = bookmark_list(user);
+        const char *md5;
+
+        xs_list_foreach(l, md5) {
+            xs *post = NULL;
+
+            if (valid_status(object_get_by_md5(md5, &post))) {
+                const char *id = xs_dict_get(post, "id");
+
+                if (xs_type(id) == XSTYPE_STRING)
+                    fprintf(f, "%s\n", id);
+            }
+        }
+
+        fclose(f);
+    }
+    else
+        snac_log(user, xs_fmt("Cannot create file %s", fn));
+
+    xs_free(fn);
+    fn = xs_fmt("%s/export/blocked_accounts.csv", user->basedir);
+    if ((f = fopen(fn, "w")) != NULL) {
+        snac_log(user, xs_fmt("Creating %s...", fn));
+
+        xs *l = muted_list(user);
+        const char *actor;
+
+        xs_list_foreach(l, actor) {
+            xs *uid = NULL;
+
+            webfinger_request_fake(actor, NULL, &uid);
+            fprintf(f, "%s\n", uid);
+        }
+
+        fclose(f);
+    }
+    else
+        snac_log(user, xs_fmt("Cannot create file %s", fn));
+
+    xs_free(fn);
+    fn = xs_fmt("%s/export/lists.csv", user->basedir);
+    if ((f = fopen(fn, "w")) != NULL) {
+        snac_log(user, xs_fmt("Creating %s...", fn));
+
+        xs *lol = list_maint(user, NULL, 0);
+        const xs_list *li;
+
+        xs_list_foreach(lol, li) {
+            const char *lid = xs_list_get(li, 0);
+            const char *ltitle = xs_list_get(li, 1);
+
+            xs *actors = list_content(user, lid, NULL, 0);
+            const char *md5;
+
+            xs_list_foreach(actors, md5) {
+                xs *actor = NULL;
+
+                if (valid_status(object_get_by_md5(md5, &actor))) {
+                    const char *id = xs_dict_get(actor, "id");
+                    xs *uid = NULL;
+
+                    webfinger_request_fake(id, NULL, &uid);
+                    fprintf(f, "%s,%s\n", ltitle, uid);
+                }
+            }
+        }
+
+        fclose(f);
+    }
+    else
+        snac_log(user, xs_fmt("Cannot create file %s", fn));
+
+    xs_free(fn);
+    fn = xs_fmt("%s/export/following_accounts.csv", user->basedir);
+    if ((f = fopen(fn, "w")) != NULL) {
+        snac_log(user, xs_fmt("Creating %s...", fn));
+
+        fprintf(f, "Account address,Show boosts,Notify on new posts,Languages\n");
+
+        xs *fwing = following_list(user);
+        const char *actor;
+
+        xs_list_foreach(fwing, actor) {
+            xs *uid = NULL;
+
+            webfinger_request_fake(actor, NULL, &uid);
+            fprintf(f, "%s,%s,false,\n", uid, limited(user, actor, 0) ? "false" : "true");
+        }
+
+        fclose(f);
+    }
+    else
+        snac_log(user, xs_fmt("Cannot create file %s", fn));
+}
+
+
+void import_blocked_accounts_csv(snac *user, const char *ifn)
+/* imports a Mastodon CSV file of blocked accounts */
+{
+    FILE *f;
+    xs *l = xs_split(ifn, "/");
+    xs *fn = xs_fmt("%s/import/%s", user->basedir, xs_list_get(l, -1));
+
+    if ((f = fopen(fn, "r")) != NULL) {
+        snac_log(user, xs_fmt("Importing from %s...", fn));
+
+        while (!feof(f)) {
+            xs *l = xs_strip_i(xs_readline(f));
+
+            if (*l && strchr(l, '@') != NULL) {
+                xs *url = NULL;
+                xs *uid = NULL;
+
+                if (valid_status(webfinger_request(l, &url, &uid))) {
+                    if (is_muted(user, url))
+                        snac_log(user, xs_fmt("Actor %s already MUTEd", url));
+                    else {
+                        mute(user, url);
+                        snac_log(user, xs_fmt("MUTEd actor %s", url));
+                    }
+                }
+                else
+                    snac_log(user, xs_fmt("Webfinger error for account %s", l));
+            }
+        }
+
+        fclose(f);
+    }
+    else
+        snac_log(user, xs_fmt("Cannot open file %s", fn));
+}
+
+
+void import_following_accounts_csv(snac *user, const char *ifn)
+/* imports a Mastodon CSV file of accounts to follow */
+{
+    FILE *f;
+    xs *l = xs_split(ifn, "/");
+    xs *fn = xs_fmt("%s/import/%s", user->basedir, xs_list_get(l, -1));
+
+    if ((f = fopen(fn, "r")) != NULL) {
+        snac_log(user, xs_fmt("Importing from %s...", fn));
+
+        while (!feof(f)) {
+            xs *l = xs_strip_i(xs_readline(f));
+
+            if (*l) {
+                xs *l2 = xs_split(l, ",");
+                const char *acct = xs_list_get(l2, 0);
+                const char *show = xs_list_get(l2, 1);
+
+                if (acct) {
+                    /* not a valid account? skip (probably the CSV header) */
+                    if (strchr(acct, '@') == NULL)
+                        continue;
+
+                    xs *msg = msg_follow(user, acct);
+
+                    if (msg == NULL) {
+                        snac_log(user, xs_fmt("Cannot follow %s -- server down?", acct));
+                        continue;
+                    }
+
+                    const char *actor = xs_dict_get(msg, "object");
+
+                    if (following_check(user, actor))
+                        snac_log(user, xs_fmt("Actor %s already followed", actor));
+                    else {
+                        following_add(user, actor, msg);
+
+                        enqueue_output_by_actor(user, msg, actor, 0);
+
+                        snac_log(user, xs_fmt("Following %s", actor));
+                    }
+
+                    if (show && strcmp(show, "false") == 0) {
+                        limit(user, actor);
+                        snac_log(user, xs_fmt("Limiting boosts from actor %s", actor));
+                    }
+                    else {
+                        unlimit(user, actor);
+                        snac_log(user, xs_fmt("Unlimiting boosts from actor %s", actor));
+                    }
+                }
+            }
+        }
+
+        fclose(f);
+    }
+    else
+        snac_log(user, xs_fmt("Cannot open file %s", fn));
+}
+
+
+void import_list_csv(snac *user, const char *ifn)
+/* imports a Mastodon CSV file list */
+{
+    FILE *f;
+    xs *l = xs_split(ifn, "/");
+    xs *fn = xs_fmt("%s/import/%s", user->basedir, xs_list_get(l, -1));
+
+    if ((f = fopen(fn, "r")) != NULL) {
+        snac_log(user, xs_fmt("Importing from %s...", fn));
+
+        while (!feof(f)) {
+            xs *l = xs_strip_i(xs_readline(f));
+
+            if (*l) {
+                xs *l2 = xs_split(l, ",");
+                const char *lname = xs_list_get(l2, 0);
+                const char *acct  = xs_list_get(l2, 1);
+
+                if (lname && acct) {
+                    /* create the list */
+                    xs *list_id = list_maint(user, lname, 1);
+
+                    xs *url = NULL;
+                    xs *uid = NULL;
+
+                    if (valid_status(webfinger_request(acct, &url, &uid))) {
+                        xs *actor_md5 = xs_md5_hex(url, strlen(url));
+
+                        list_content(user, list_id, actor_md5, 1);
+                        snac_log(user, xs_fmt("Added %s to list %s", url, lname));
+
+                        if (!following_check(user, url)) {
+                            xs *msg = msg_follow(user, url);
+
+                            if (msg == NULL) {
+                                snac_log(user, xs_fmt("Cannot follow %s -- server down?", acct));
+                                continue;
+                            }
+
+                            following_add(user, url, msg);
+
+                            enqueue_output_by_actor(user, msg, url, 0);
+
+                            snac_log(user, xs_fmt("Following %s", url));
+                        }
+                    }
+                    else
+                        snac_log(user, xs_fmt("Webfinger error while adding %s to list %s", acct, lname));
+                }
+            }
+        }
+
+        fclose(f);
+    }
+    else
+        snac_log(user, xs_fmt("Cannot open file %s", fn));
+}
+
+
+void import_csv(snac *user)
+/* import CSV files from Mastodon */
+{
+    FILE *f;
+
+    import_blocked_accounts_csv(user, "blocked_accounts.csv");
+
+    import_following_accounts_csv(user, "following_accounts.csv");
+
+    import_list_csv(user, "lists.csv");
+
+    xs *fn = xs_fmt("%s/import/bookmarks.csv", user->basedir);
+    if ((f = fopen(fn, "r")) != NULL) {
+        snac_log(user, xs_fmt("Importing from %s...", fn));
+
+        while (!feof(f)) {
+            xs *l = xs_strip_i(xs_readline(f));
+
+            if (*l) {
+                xs *post = NULL;
+
+                if (!object_get(l, &post)) {
+                    if (!valid_status(activitypub_request(user, l, &post))) {
+                        snac_log(user, xs_fmt("Error getting object %s for bookmarking", l));
+                        continue;
+                    }
+                }
+
+                if (post == NULL)
+                    continue;
+
+                /* request the actor that created the post */
+                const char *actor = get_atto(post);
+
+                if (xs_type(actor) == XSTYPE_STRING)
+                    actor_request(user, actor, NULL);
+
+                object_add_ow(l, post);
+                timeline_add(user, l, post);
+
+                bookmark(user, l);
+
+                snac_log(user, xs_fmt("Bookmarked %s", l));
+            }
+        }
+
+        fclose(f);
+    }
+    else
+        snac_log(user, xs_fmt("Cannot open file %s", fn));
+}
