@@ -17,7 +17,7 @@ xs_str *xs_url_dec(const char *str)
 
     while (*str) {
         if (*str == '%') {
-            int i;
+            unsigned int i;
 
             if (sscanf(str + 1, "%02x", &i) == 1) {
                 unsigned char uc = i;
@@ -53,7 +53,8 @@ xs_dict *xs_url_vars(const char *str)
         const xs_val *v;
 
         xs_list_foreach(args, v) {
-            xs *kv = xs_split_n(v, "=", 1);
+            xs *dv = xs_url_dec(v);
+            xs *kv = xs_split_n(dv, "=", 1);
 
             if (xs_list_len(kv) == 2) {
                 const char *key = xs_list_get(kv, 0);
@@ -105,13 +106,13 @@ xs_dict *xs_multipart_form_data(const char *payload, int p_size, const char *hea
         if (xs_list_len(l1) != 2)
             return NULL;
 
-        boundary = xs_dup(xs_list_get(l1, 1));
+        xs *t_boundary = xs_dup(xs_list_get(l1, 1));
 
         /* Tokodon sends the boundary header with double quotes surrounded */
-        if (xs_between("\"", boundary, "\"") != 0)
-            boundary = xs_strip_chars_i(boundary, "\"");
+        if (xs_between("\"", t_boundary, "\"") != 0)
+            t_boundary = xs_strip_chars_i(t_boundary, "\"");
 
-        boundary = xs_fmt("--%s", boundary);
+        boundary = xs_fmt("--%s", t_boundary);
     }
 
     bsz = strlen(boundary);
@@ -152,7 +153,7 @@ xs_dict *xs_multipart_form_data(const char *payload, int p_size, const char *hea
             memcpy(s1, p, q - p);
             s1[q - p] = '\0';
 
-            if (xs_startswith(s1, "Content-Disposition")) {
+            if (xs_startswith(s1, "Content-Disposition") || xs_startswith(s1, "content-disposition")) {
                 /* split by " like a primitive man */
                 l1 = xs_split(s1, "\"");
 
@@ -166,7 +167,7 @@ xs_dict *xs_multipart_form_data(const char *payload, int p_size, const char *hea
                 }
             }
             else
-            if (xs_startswith(s1, "Content-Type")) {
+            if (xs_startswith(s1, "Content-Type") || xs_startswith(s1, "content-type")) {
                 l1 = xs_split(s1, ":");
 
                 if (xs_list_len(l1) >= 2) {
@@ -195,7 +196,7 @@ xs_dict *xs_multipart_form_data(const char *payload, int p_size, const char *hea
         if (fn != NULL) {
             /* p_var value is a list */
             /* if filename has no extension and content-type is image, attach extension to the filename */
-            if (strchr(fn, '.') == NULL && xs_startswith(ct, "image/")) {
+            if (strchr(fn, '.') == NULL && ct && xs_startswith(ct, "image/")) {
                 char *ext = strchr(ct, '/');
                 ext++;
                 fn = xs_str_cat(xs_str_new(""), fn, ".", ext);
